/*!
    \file
    \brief Defines the class GLWidget.
*/


// TEST
#include <GL/glew.h>
#include <GL/glut.h>

#include <cuda_runtime.h>
#include <cutil_inline.h>
#include <cutil_gl_inline.h>
#include <cuda_gl_interop.h>

#include "SobelFilter_kernels.h"
//


#include "GLWidget.h"


#include <QtOpenGL>
#include <QDebug>
#include <QFileInfo>
#include <QTimer>
#include <QImage>
#include <QMovie>
#include <QString>




/*!
    \brief Constructor with a QString for the QImage.
    \param parent      The parent widget.
    \param qsImagePath The path to the image file as a QString.
 */
GLWidget::GLWidget(QWidget *parent, const QString &qsImagePath)
        : QGLWidget(parent)
{
    QImage img(qsImagePath);

    if (!setImagePath(qsImagePath) || !setImage(img))
    {
        qDebug() << "Aborting.";
        abort();
    }

    QTimer *timer = new QTimer();
    connect(timer, SIGNAL(timeout()), this, SLOT(updateGL()));
    timer->start(200);

    setFixedSize(m_pQImage.width(), m_pQImage.height());

    // TMP !!
    //m_pQMovie = new QMovie(qsImagePath);
    //AssertValidConnection(connect(m_pQMovie, SIGNAL(updated(const QRect&)), this, SLOT(slotUpdate())));
    //m_pQMovie->start();
}


/*!
    \brief Constructor with a QImage directly.
    \param parent      The parent widget.
    \param qiImage     The image for the texture as a QImage.
 */
GLWidget::GLWidget(QWidget *parent, const QImage &qiImage)
        : QGLWidget(parent)
        , m_pQImage(qiImage)
{
    if (!setImage(qiImage))
    {
        qDebug() << "Aborting.";
        abort();
    }

    setFixedSize(m_pQImage.width(), m_pQImage.height());
}


//! Common destructor.
GLWidget::~GLWidget()
{
    makeCurrent();

    glDeleteBuffers (1, &pbo_buffer);
    glDeleteTextures(1, &texid);
    glDeleteTextures(1, &m_pTexture);

    pbo_buffer = 0;
    texid = 0;
    m_pTexture = 0;
    
}


/*!
    \brief Sets OpenGL parameters and starts the texture loading routine \ref loadGLTexture().

    This function is called once automatically, when the widget is made visible for the first time.
 */
void GLWidget::initializeGL()
{
    qDebug() << "initializeGL()";
    makeCurrent();

    // be sure that the QImage is already loaded correctly before calling this function!
    m_pQImage = QGLWidget::convertToGLFormat(m_pQImage);

    cudaGLSetGLDevice (cutGetMaxGflopsDeviceId() );

/*
    int c=1;
    char* dummy = "";
    glutInit( &c, &dummy );
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
*/

    glewInit();

    /*
    if (!glewIsSupported( "GL_VERSION_2_0 GL_ARB_fragment_program GL_EXT_framebuffer_object" )) {
        fprintf(stderr, "Error: failed to get minimal extensions for demo\n");
        fprintf(stderr, "This sample requires:\n");
        fprintf(stderr, "  OpenGL version 2.0\n");
        fprintf(stderr, "  GL_ARB_fragment_program\n");
        fprintf(stderr, "  GL_EXT_framebuffer_object\n");
        cudaThreadExit();
        exit(-1);
    }
	if (!glewIsSupported( "GL_VERSION_1_5 GL_ARB_vertex_buffer_object GL_ARB_pixel_buffer_object" )) {
		fprintf(stderr, "Error: failed to get minimal extensions for demo\n");
		fprintf(stderr, "This sample requires:\n");
		fprintf(stderr, "  OpenGL version 1.5\n");
		fprintf(stderr, "  GL_ARB_vertex_buffer_object\n");
		fprintf(stderr, "  GL_ARB_pixel_buffer_object\n");
        cudaThreadExit();
		exit(-1);
	}
    */

    unsigned int w, h;
    char data[1024];
    strncpy(data, getImagePath().toLatin1(), sizeof(data) - 1);
    const char* image_filename = data;

    if (cutLoadPPM4ub(image_filename, &pixels, &w, &h) != CUTTrue) {
        printf("Failed to load image file: %s\n", image_filename);
        exit(-1);
    }


    GLint bsize;

    //pixels = m_pQImage.bits();
    setupTexture(m_pQImage.width(), m_pQImage.height(), pixels, 4);

    //memset(pixels, 0x0, 4 * sizeof(Pixel) * m_pQImage.width() * m_pQImage.height());

    // use OpenGL Path
    //glEnable(GL_TEXTURE_2D);
    //glAlphaFunc(GL_GREATER, 0.1f);
    //glEnable(GL_ALPHA_TEST);

    glGenBuffers(1, &pbo_buffer);
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, pbo_buffer); 
    glBufferData(GL_PIXEL_UNPACK_BUFFER, 
                    4 * sizeof(Pixel) * m_pQImage.width() * m_pQImage.height(), 
                    pixels, GL_STREAM_DRAW);

    glGetBufferParameteriv(GL_PIXEL_UNPACK_BUFFER, GL_BUFFER_SIZE, &bsize); 
    if ((GLuint)bsize != (4 * sizeof(Pixel) * m_pQImage.width() * m_pQImage.height()))
    {
        printf("Buffer object (%d) has incorrect size (%d).\n", (unsigned)pbo_buffer, (unsigned)bsize);
        cudaThreadExit();
        exit(-1);
    }

    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);

    // register this buffer object with CUDA
    cutilSafeCall(cudaGraphicsGLRegisterBuffer(&cuda_pbo_resource, pbo_buffer, cudaGraphicsMapFlagsWriteDiscard));	


    glGenTextures  (1, &m_pTexture);
    glBindTexture  (GL_TEXTURE_2D, m_pTexture);
    glTexImage2D   (GL_TEXTURE_2D, 0, GL_RGBA, m_pQImage.width(), m_pQImage.height(), 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
    //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glBindTexture  (GL_TEXTURE_2D, 0);

    glPixelStorei  (GL_UNPACK_ALIGNMENT, 1);
    glPixelStorei  (GL_PACK_ALIGNMENT, 1);

    //glDisable(GL_ALPHA_TEST);
    //glDisable(GL_TEXTURE_2D);

}


/*!
    \brief Reimplementation of the function to paint the (Q)GLWidget.
    
    This function is called automatically every time the widget has to be shown or is changed in any way.
 */
void GLWidget::paintGL()
{
    qDebug() << "paintGL()";

    makeCurrent();

    //glMatrixMode(GL_PROJECTION);
    //glLoadIdentity();
    //glMatrixMode(GL_MODELVIEW);

    // Sobel operation
    Pixel *data = NULL;

    // map PBO to get CUDA device pointer
	cutilSafeCall(cudaGraphicsMapResources(1, &cuda_pbo_resource, 0));
    size_t num_bytes; 
    cutilSafeCall(cudaGraphicsResourceGetMappedPointer((void **)&data, &num_bytes,  
						       cuda_pbo_resource));
    qDebug() << QString("CUDA mapped PBO: May access %1 bytes").arg(num_bytes);
	
	//sobelFilter(data, m_pQImage.width(), m_pQImage.height(), SOBELDISPLAY_IMAGE, 1.0f);
    cutilSafeCall(cudaGraphicsUnmapResources(1, &cuda_pbo_resource, 0));

    glClear(GL_COLOR_BUFFER_BIT);

    glBindTexture(GL_TEXTURE_2D, m_pTexture);
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, pbo_buffer);
    glTexImage2D(GL_TEXTURE_2D, 0, 0, 0, m_pQImage.width(), m_pQImage.height(), 
                   GL_LUMINANCE, GL_UNSIGNED_BYTE, NULL /*= OFFSET*/);
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);

    glDisable(GL_DEPTH_TEST);
    glEnable(GL_TEXTURE_2D);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        
    glBegin(GL_QUADS);
    //glVertex2f(0, 0); glTexCoord2f(0, 0);
    //glVertex2f(0, 1); glTexCoord2f(1, 0);
    //glVertex2f(1, 1); glTexCoord2f(1, 1);
    //glVertex2f(1, 0); glTexCoord2f(0, 1);
        glTexCoord2f(0.0f, 0.0f);
        glVertex2f(-1.0f, -1.0f);
        glTexCoord2f(1.0f, 0.0f);
        glVertex2f(1.0f, -1.0f);
        glTexCoord2f(1.0f, 1.0f);
        glVertex2f(1.0f, 1.0f);
        glTexCoord2f(0.0f, 1.0f);
        glVertex2f(-1.0f, 1.0f);
    glEnd();

    //glDisable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, 0);

    swapBuffers();
}


/*
    \brief Reimplementation of the resizing function for the (Q)GLWidget.

    This function is called automatically, when the widget gets resized.

    \param width  The new width for resizing the widget in pixels.
    \param height The new height for resizing the widget in pixels.
 */
void GLWidget::resizeGL(int iWidth, int iHeight)
{
    qDebug() << "resizeGL()";
/*
    makeCurrent();
    int iSide = qMin(iWidth, iHeight);
    glViewport((iWidth - iSide) / 2, (iHeight - iSide) / 2, iSide, iSide);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-0.5, +0.5, +0.5, -0.5, 4.0, 15.0);
    glMatrixMode(GL_MODELVIEW);

    setFixedSize(iWidth, iHeight);
*/
}


//! Convert the image to an OpenGL texture and bind it to the texture quad.
void GLWidget::loadGLTexture()
{
    // very important!
    m_pQImage = QGLWidget::convertToGLFormat(m_pQImage);

    makeCurrent();

    glEnable(GL_TEXTURE_2D);
    glAlphaFunc(GL_GREATER, 0.1f);
    glEnable(GL_ALPHA_TEST);

    glGenTextures  (1, &m_pTexture);
    glBindTexture  (GL_TEXTURE_2D, m_pTexture);
    glTexImage2D   (GL_TEXTURE_2D, 0, 4, m_pQImage.width(), m_pQImage.height(), 0, GL_RGBA, GL_UNSIGNED_BYTE, m_pQImage.bits());
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

    glDisable(GL_ALPHA_TEST);
    glDisable(GL_TEXTURE_2D);
}


//! Getter for the path of the image as a QString.
const QString GLWidget::getImagePath()
{
    return m_pQSImagePath;
}


//! Getter for the current image as a Qimage.
const QImage  GLWidget::getImage()
{
    return m_pQImage;
}


/*!
    \brief  Sets the image path for \ref m_pQSImagePath if the new file really exists.
    \return Returns true on success, otherwise returns false.
 */
bool GLWidget::setImagePath(const QString &qsImagePath)
{
    if (qsImagePath.isEmpty())
    {
        qDebug() << "The Image path is empty!";

        return false;
    }

    else if (!QFileInfo(qsImagePath).exists())
    {
        qDebug() << "The file '" << qsImagePath << "' does not exist!";

        return false;
    }

    else
    {
        m_pQSImagePath = qsImagePath;

        return true;
    }
}


/*!
    \brief  Sets the image for \ref m_pQImage if the new image is not a null image.
    \return Returns true on success, otherwise returns false.
 */
bool GLWidget::setImage(const QImage &qiImage)
{
    if (qiImage.isNull())
    {
        qDebug() << "The image is a null image!";

        return false;
    }

    else
    {
        m_pQImage = qiImage;

        return true;
    }

}


/*!
    \brief  Tries to update the image for \ref m_pQImage from a given path and display it.
    \param  qsImagePath The new path for the image in \ref m_pQImage.
    \return Returns true on success, otherwise returns false.
 */
bool GLWidget::slotUpdateImage(const QString &qsImagePath)
{
    if (!setImagePath(qsImagePath))
        return false;

    else
    {
        if (!m_pQImage.load(m_pQSImagePath))
        {
            qDebug() << "Could not load image: '" << m_pQSImagePath << "'!";

            return false;
        }

        else
        {
            // very important! prevents memory leaks.
            glDeleteTextures(1, &m_pTexture);
            m_pTexture = 0;

            loadGLTexture();
            updateGL();

            return true;
        }
    }
}


/*!
    \brief  Tries to update the image for \ref m_pQImage and display it.
    \param  qiImage The new image for \ref m_pQImage.
    \return Returns true on success, otherwise returns false.
 */
bool GLWidget::slotUpdateImage(const QImage &qiImage)
{
    if (!setImage(qiImage))
        return false;

    else
    {
        // very important! prevents memory leaks.
        glDeleteTextures(1, &m_pTexture);
        m_pTexture = 0;

        loadGLTexture();
        updateGL();

        return true;
    }
}


//! Public slot to update the OpenGL texture with the current image from \ref m_pQMovie.
void GLWidget::slotUpdate()
{
    //if(!slotUpdateImage(m_pQMovie->currentImage()))
    //    qDebug() << "slotUpdateImage failed!";

    updateGL();
}


//! Public slot to resize the current OpenGL viewport to the size of \ref m_pQImage.
void GLWidget::slotResize()
{
    resizeGL(m_pQImage.width(), m_pQImage.height());
}




//////////////////////
//////////////////////
//////////////////////

















